# -*- coding: utf-8 -*-
import time
import sys
from collections import Iterable
import six


def floatformatter(*args, sig=6, **kwargs):
    return "{" + "0:.{}g".format(sig) + "}"


def float_to_str_sig(value, *args, sig: int=6, atol: float=1e-7, **kwargs):
    """
    Returns a string representation of a floating point number, with
    given significant digits.

    Parameters
    ----------
    value : float or a sequence of floats
        A single value, or an iterable.

    sig : int
        Number of significant digits.

    atol : float
        Floating point tolerance. Values smaller than this 
        in the absolute sense are treated as zero.

    Returns
    -------
    string or a sequence of strings
        String representation of the provided input.
    """
    if not issequence(value):
        if atol is not None:
            if abs(value) < atol:
                value = 0.0
        return floatformatter(sig=sig).format(value)
    else:
        try:
            import numpy as np
        except ImportError:
            raise ImportError("You need numpy for this.")
        value = np.array(value)
        if atol is not None:
            inds = np.where(np.abs(value) < atol)[0]
            value[inds] = 0.0
        formatter = floatformatter(sig=sig)
        def f(v): return formatter.format(v)
        return list(map(f, value))


def timer(fnc):
    def inner(*args, **kwargs):
        t0 = time.time()
        fnc(*args, **kwargs)
        t1 = time.time()
        return t1-t0
    return inner


def SuppressedFunction(fnc):
    """Decorator that wraps a function to suppress 
    it's calls to print()"""
    def inner(*arg):
        original_stdout = sys.stdout
        sys.stdout = None
        res = fnc(*arg)
        sys.stdout = original_stdout
        return res
    return inner


def alphabet(abctype: str = 'latin', **kwargs):
    if abctype in ('ord', 'o'):
        start = kwargs.pop('start', 0)
    elif abctype in ('latin', 'l'):
        start = ord(kwargs.pop('start', 'a'))
    elif abctype == 'u':
        start = ord(kwargs.pop('start', '\u0000'))
    elif abctype in ('greek', 'g'):
        start = ord(kwargs.pop('start', '\u03b1'))
    while True:
        yield chr(start)
        start += 1


def ordrange(N: int = 1, **kwargs):
    start = kwargs.pop('start', 0)
    if isinstance(start, str):
        start = ord(start)
    stop = kwargs.pop('stop', None)
    if stop is None or stop == start:
        stop = start + N
    return [chr(c) for c in range(start, stop)]


def latinrange(N: int = 1, **kwargs):
    start = kwargs.pop('start', 97)
    stop = kwargs.pop('stop', None)
    return ordrange(N, start=start, stop=stop)


def urange(N: int = 1, **kwargs):
    start = kwargs.pop('start', '\u0000')
    stop = kwargs.pop('stop', None)
    if stop is None:
        stop = start
    return ordrange(N, start=ord(start), stop=ord(stop))


def greekrange(N: int = 1):
    return urange(N, start='\u03b1')


def arabicrange(N: int = 1, **kwargs):
    start = kwargs.pop('start', 0)
    stop = kwargs.pop('stop', None)
    if stop is None or stop == start:
        stop = start + N
    return [str(c) for c in range(start, stop)]


def issequence(arg):
    """
    A sequence is an iterable, but not any kind of string.
    """
    return (
        isinstance(arg, Iterable)
        and not isinstance(arg, six.string_types)
    )


if __name__ == '__main__':

    lrange = latinrange(5, start='i')
    grange = greekrange(5)
    orange = ordrange(5)
    arange = arabicrange(5, start=1)

    abc = alphabet('latin', start='i')
    for i in range(6):
        print(next(abc))

    abc = alphabet(start='u')
    for i in range(6):
        print(next(abc))
        
    abc = alphabet('u', start='\x03')
    for i in range(4):
        print(next(abc))
    
    abc = alphabet('u', start='\x03')
    pokerstr = [next(abc) for _ in range(4)]
    print("{} {} {} {}".format(*pokerstr))



# -*- coding: utf-8 -*-
from dewloosh.core.types.numberedentity import UniqueNumbered
import numpy as np


class NumberedList(list):
    """
    Makes it possible to loop over a list of items with discontinous numbering.
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.Sequence = None
        self.SortedLoop = False
        self.counter = None
        return

    def build_sequence(self):
        state = self.SortedLoop
        self.SortedLoop = False
        self.Sequence = np.zeros(len(self), dtype=np.int32)
        numbers = np.array([item.Num for item in self], dtype=np.int32)
        self.Sequence = np.argsort(numbers)
        self.SortedLoop = state
        return

    def __iter__(self):
        if (self.SortedLoop == True) and (self.Sequence is None):
            self.build_sequence()
        self.counter = -1
        return self

    def __next__(self):
        if self.counter < len(self)-1:
            self.counter += 1
            if self.SortedLoop:
                return self[self.Sequence[self.counter]]
            else:
                return self[self.counter]
        else:
            self.counter = -1
            raise StopIteration

    def append(self, item):
        if not hasattr(item, 'Num'):
            raise TypeError("Item doesn't have attribute 'Num'.")
        super().append(item)
        return


if __name__ == '__main__':

    item1 = UniqueNumbered(num=2)
    item2 = UniqueNumbered(num=1)
    item3 = UniqueNumbered(num=-5)
    item4 = UniqueNumbered(num=100)
    item5 = UniqueNumbered(num=-1)

    l = NumberedList()
    l.append(item1)
    l.append(item2)
    l.append(item3)
    l.append(item4)
    l.append(item5)

    print('\nfirst print')
    l.SortedLoop = True
    for item in l:
        print(item.Num)

    print('\nsecond print')
    l.SortedLoop = False
    for item in l:
        print(item.Num)



# -*- coding: utf-8 -*-
from types import FunctionType
from abc import abstractmethod


def attributor(*attrs: str) -> FunctionType:
    """
    It renders a decorator a default behaviour. If a decorator
    is called with a None argument, it returns the attribute, otherwise it 
    returns the decorated function.
    """
    abstract = '__isabstractmethod__' in attrs
    if abstract:
        attributes = [x for x in attrs if x != '__isabstractmethod__']
    else:
        attributes = attrs

    def decorator(fnc):
        if fnc is None:
            return attrs
        else:
            for attr in attributes:
                setattr(fnc, attr, True)
        if abstract:
            return abstractmethod(fnc)
        return fnc
    return decorator


if __name__ == '__main__':

    axiom = attributor('__isaxiom__')
    abstractaxiom = attributor('__isaxiom__', '__isabstractmethod__')

    @axiom
    def foo(a, b):
        return 'an axiom'



# -*- coding: utf-8 -*-
from functools import reduce


def chain1(*funcs):
    def foo(*args, **kwargs):
        res = funcs[-1](*args, **kwargs)
        for func in funcs[:-1]:
            res = func(res)
        return res
    return foo


def chain2(*funcs):
    def bop(f1, f2):
        def foo(*args, **kwargs):
            return f1(f2(*args, **kwargs))
        return foo
    return reduce(bop, funcs)


def chain3(*funcs):
    def foo(*args, **kwargs):
        if len(funcs) == 1:
            return funcs[0](*args, **kwargs)
        else:
            return funcs[0](chain3(*funcs[1:])(*args, **kwargs))
    return foo


if __name__ == '__main__':

    def lambdagen1d(start, stop): return map(
        lambda p: lambda x: x**p, range(start, stop))

    def f(x): return x**2
    def g(x): return x + 1

    h = chain1(f, g)
    print(h(2))
    print(h(3))

    print("------\n")

    m = chain2(f, g)
    print(m(2))
    print(m(3))

    print("------\n")

    k = chain3(f, g)
    print(k(2))
    print(k(3))

    print("------\n")

    for f in lambdagen1d(0, 4):
        print(f(2))



# -*- coding: utf-8 -*-
from dewloosh.core.types.defaultdict import OrderedDictCollection, \
    parsedicts
from dewloosh.core.tools import issequence


class Library(OrderedDictCollection):

    def __init__(self, *args, parent=None, root=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent
        self._root = root
        self._locked=False
    
    @property        
    def locked(self):
        return self._locked
    
    @locked.setter 
    def locked(self, value):
        assert isinstance(value, bool)
        self._locked = value
    
    def lock(self):
        self._locked=True
        
    def unlock(self):
        self._locked=False

    def root(self):
        if self.parent is None:
            return self
        else:
            if self._root is not None:
                return self._root
            else:
                return self.parent.root()

    def is_root(self):
        return self.parent is None

    def containers(self, *args, inclusive=False, deep=True, dtype=None, **kwargs):
        dtype = Library if dtype is None else dtype
        return parsedicts(self, inclusive=inclusive, dtype=dtype, deep=deep)

    def __missing__(self, key):
        if self._locked:
            raise KeyError("Missing key : {}".format(key))
        else:
            return super().__missing__(key)

    def __join_parent__(self, parent: 'Library'):
        self.parent = parent
        self._root = parent.root()

    def __repr__(self):
        return dict.__repr__(self)

    def __setitem__(self, key, value):
        try:
            if issequence(key):
                if not key[0] in self:
                    d = self.__missing__(key[0])
                else:
                    d = self[key[0]]
                if len(key) > 1:
                    d.__setitem__(key[1:], value)
                else:
                    self[key[0]] = value
            else:
                if isinstance(value, Library):
                    value.__join_parent__(self)
                return super().__setitem__(key, value)
        except AttributeError:
            raise RuntimeError("Target is of type '{}', which is not \
                               a container.".format(type(d)))
        except KeyError:
            return self.__missing__(key)

    def default_factory(self):
        cls = type(self)
        return cls(parent=self, root=self.root())


if __name__ == '__main__':

    h = Library()
    h['a', 'b', 'c', 'e'] = 1
    #h['a']['b']['c']['e'] = 1
    h['a']['b']['d'] = 2

    b = h['a', 'b']
    b['e'] = 3
    b['f'] = 1, 2, 3



# -*- coding: utf-8 -*-
from abc import ABC
import uuid

__all__ = ['Numbered', 'Unique']


class Numbered(ABC):

    def __init__(self, *args, num=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.num = num

    def __enter__(self):
        return self

    def __exit__(self, *a):
        pass

    def __eq__(self, other):
        return self.num == other.num


class Unique(ABC):

    def __init__(self, *args, key=None, **kwargs):
        super().__init__(*args, **kwargs)
        if 'UUID' in kwargs:
            UUID = kwargs.pop('UUID')
        elif 'UID' in kwargs:
            UUID = kwargs.pop('UID')
        else:
            UUID = uuid.uuid4()

        if 'GUID' in kwargs:
            GUID = kwargs.pop('GUID')
        elif 'GID' in kwargs:
            GUID = kwargs.pop('GID')
        else:
            GUID = None

        if 'ID' in kwargs:
            ID = kwargs.pop('ID')
        elif 'LID' in kwargs:
            ID = kwargs.pop('LID')
        else:
            ID = None

        self._UUID = UUID
        self._ID = ID
        self._GUID = GUID
        self._key = key

    def __eq__(self, other):
        return isinstance(other, Unique) and self.UUID == other.UUID

    @property
    def to_hex(self):
        return self.UUID.hex

    @property
    def to_int(self):
        return self.UUID.int

    @property
    def GUID(self):
        return self._GUID

    @GUID.setter
    def GUID(self, value):
        self._GUID = value

    @property
    def GID(self):
        return self._GUID

    @GID.setter
    def GID(self, value):
        self._GUID = value

    @property
    def ID(self):
        return self._ID

    @ID.setter
    def ID(self, value):
        self._ID = value

    @property
    def LID(self):
        return self._ID

    @LID.setter
    def LID(self, value):
        self._ID = value

    @property
    def UUID(self):
        return self._UUID

    @property
    def UID(self):
        return self._UUID

    def __hash__(self):
        return self.to_int()

    @property
    def key(self):
        return self._key if self._key is not None else self._UUID

    @key.setter
    def key(self, value):
        self._key = value

    def __str__(self):
        return str(self.key)


class UniqueNumbered(Numbered, Unique):
    ...


if __name__ == '__main__':

    u1 = UniqueNumbered()
    u2 = UniqueNumbered(key='somekey')



# -*- coding: utf-8 -*-
import os


def desktopfolder():
    return os.path.join(os.path.join(os.environ['USERPROFILE']), 'Desktop')


def dir_name_ext(fullpath: str):
    dirpath, filename = os.path.split(fullpath)
    name, ext = filename.split(".")
    return dirpath, name, ext


def fullpath_to_dir(fullpath: str):
    return dir_name_ext(fullpath)[0]


def fullpath_to_name(fullpath: str):
    return dir_name_ext(fullpath)[1]


def fullpath_to_ext(fullpath: str):
    return dir_name_ext(fullpath)[2]
